import os
from datetime import datetime

import numpy as np
import pandas as pd

from mne import Epochs
from mne.utils import logger


def update_log(log_file_path: str, epochs: Epochs, notes: str) -> pd.DataFrame:
    """
    Documents the changes during preprocessing for an Epochs object.
    Custom description can be added with the notes argument.
    The file path has to contain the file extension too (.csv).

    Parameters
    ----------
    log_file_path
    epochs
    notes

    Returns
    ----------
    log data
    """
    fid = epochs.info["temp"]
    dropped_epochs_marker = ["FASTER", "USER", "AUTOREJECT"]
    n_bad_epochs = len(
        [drop for drop in epochs.drop_log if np.isin(dropped_epochs_marker, drop).any()]
    )

    log = pd.DataFrame(
        {
            "fid": [fid],
            "highpass": [epochs.info["highpass"]],
            "lowpass": [epochs.info["lowpass"]],
            "n_components": [np.NaN],
            "n_bad_epochs": [n_bad_epochs],
            "n_total_epochs": [len(epochs)],
            "drop_percentage": [round(epochs.drop_log_stats(), 2)],
            "stimuli": [list(epochs.event_id.keys())],
            "t_min": [epochs.tmin],
            "t_max": [epochs.tmax],
            "n_interpolated": [np.NaN],
            "average_ref_applied": [bool(epochs.info["custom_ref_applied"])],
            "baseline": [epochs.baseline if epochs.baseline else np.NaN],
            "notes": [notes],
            "date_of_update": [datetime.utcnow().isoformat()],
        }
    )

    description = epochs.info["description"]

    if description is None:
        logger.warning(
            'Failed to update parameters from epochs.info["description"], \n'
            "Returning log object, consider adding description field manually and rerunning this function.\n"
            'Formatting should match this format: "n_components: 1, interpolated: AF7, Fp2, P3, CP5, Fp1"'
        )
        return log

    if "n_components" in description:
        n_components = [x for x in description.split(",")[0] if x.isdigit()][0]
        log["n_components"].update(int(n_components))

    if "interpolated" in description:
        description_plain = description.replace(" ", "").replace(":", "")
        interpolated_channels_str = description_plain.split("interpolated")[1]
        n_interpolated = len(interpolated_channels_str.split(","))
        log["n_interpolated"].update(n_interpolated)

    if os.path.isfile(log_file_path):
        log.to_csv(log_file_path, mode="a", index=False, header=False)
    else:
        log.to_csv(log_file_path, index=False)

    return log



"""
This module contains functions to perform time-frequency analysis on EEG/MEG dataset.
"""
import os
from pathlib import Path

import yaml
import pandas as pd
from tqdm import tqdm
from mne import read_epochs
from mne.utils import logger

from meeg_tools.time_frequency import compute_power, save_to_hdf5

# file path to configuration file
CONFIGURATION_FILE_PATH = "utils/config.yml"
# file path to preprocessed epochs
EPOCHS_FILE_PATH = "/Users/weian/Downloads/epochs_asrt"
# postfix to lookup when collecting the files
EPOCHS_FILE_POSTFIX = "autoreject_ransac-epo.fif.gz"


def run_tfr_pipeline(source: Path, target: Path, conf: dict):
    def _run_tfr(epochs_tfr):
        if mode == "power":
            power = compute_power(epochs=epochs_tfr, config=conf["power"])
            power.comment = file_path
            save_to_hdf5(power=power)
        else:
            pass

    files = sorted(list(source.rglob(f"*{EPOCHS_FILE_POSTFIX}")))

    if not len(files):
        logger.warning(
            f"There are no files with the expected {EPOCHS_FILE_POSTFIX}."
            f"Doing nothing ..."
        )
        return

    pbar = tqdm(sorted(files))
    for file in pbar:
        pbar.set_description("Processing %s" % file.stem)
        fid = "_".join(str(file.stem.replace(" ", "_")).split("_")[:3])

        epochs = read_epochs(os.path.join(source, file), preload=False, verbose=0)
        epochs.load_data().pick_types(**{config["analysis"]["picks"]: True})
        # make sure average ref is applied
        if not epochs.info["custom_ref_applied"]:
            epochs.load_data().set_eeg_reference("average")

        condition_names = conf["analysis"]["conditions"]
        mode = conf["analysis"]["mode"]
        if condition_names:
            if isinstance(epochs.metadata, pd.DataFrame):
                epochs.metadata = epochs.metadata.astype(str)
                if all(name in epochs.metadata.columns for name in condition_names):
                    for condition_values, _ in epochs.metadata.groupby(condition_names):
                        query = [
                            f'{a} == "{b}"'
                            for a, b in zip(condition_names, condition_values)
                        ]
                        query_str = " & ".join(query)
                        epochs_query = epochs[query_str]

                        file_name = f"{fid}_{'_'.join(condition_values)}"
                        file_path = target / file_name
                        if file_exists(file_path=f"{file_path}_{mode}-tfr.h5"):
                            pbar.update(1)
                            continue

                        _run_tfr(epochs_tfr=epochs_query)

        else:
            file_path = target / fid
            if file_exists(f'{file_path}_{conf["analysis"]["mode"]}-tfr.h5'):
                pbar.update(1)
                continue

            _run_tfr(epochs_tfr=epochs)

    pbar.close()


def read_config(file_name: str):
    cfg = {}
    if os.path.exists(file_name):
        with open(file_name, "r") as config_file:
            cfg = yaml.safe_load(config_file)
    else:
        logger.error(f"Configuration file ({file_name}) does not exist!")
    return cfg


def file_exists(file_path: str) -> bool:
    return os.path.exists(file_path) and not config["analysis"]["overwrite"]


if __name__ == "__main__":
    logger.info(f"\nReading configuration from {CONFIGURATION_FILE_PATH} file ...\n")
    config = read_config(CONFIGURATION_FILE_PATH)
    logger.info(yaml.dump(config))
    logger.info(
        f'\nRunning EEG/MEG {config["analysis"]["mode"]} analysis pipeline for '
        f'{config["analysis"]["picks"]} channels\n'
    )

    pipeline_in = Path(EPOCHS_FILE_PATH)
    if len(config["analysis"]["conditions"]):
        conditions = "-".join(config["analysis"]["conditions"]).upper()
        pipeline_out = (
            pipeline_in.parent
            / config["analysis"]["picks"]
            / config["analysis"]["mode"]
            / conditions
        )
    else:
        pipeline_out = (
            pipeline_in.parent
            / config["analysis"]["picks"]
            / config["analysis"]["mode"]
        )
    logger.info(f'Source directory is "{pipeline_in}"')
    logger.info(f'Output files are created at "{pipeline_out}"')

    if len(config["analysis"]["conditions"]):
        logger.info(
            f'Current run computes {config["analysis"]["mode"]} '
            f'for each {", ".join(config["analysis"]["conditions"])} '
            f"found in epochs metadata."
        )

    run_tfr_pipeline(source=pipeline_in, target=pipeline_out, conf=config)



"""
This module contains functions that can be used to perform connectivity analysis
in the sensor space using MNE-Python.
https://mne.tools/mne-connectivity/stable/generated/mne_connectivity.spectral_connectivity.html
https://mne.tools/mne-connectivity/stable/auto_examples/sensor_connectivity.html#sphx-glr-auto-examples-sensor-connectivity-py
"""
import numpy as np
from matplotlib import pyplot as plt

from mne import Epochs
from mne.connectivity import spectral_connectivity
from mne.preprocessing import compute_current_source_density


def compute_connectivity(epochs: Epochs, **kwargs) -> np.ndarray:
    """
    Compute channel level connectivity matrix from Epochs instance.
    Returns the computed connectivity matrix (n_freqs, n_signals, n_signals).

    Args:
        str spectrum_mode: Valid estimation mode 'fourier' or 'multitaper'
        Epochs epochs: Epochs extracted from a Raw instance
        str method: connectivity estimation method
        int n_jobs: number of epochs to process in parallel
    :return: np.ndarray con: The computed connectivity matrix with a shape of
    (n_freqs, n_signals, n_signals).

    See Also
    --------
    For frequency-decomposition and frequency bin reference:
    https://docs.scipy.org/doc/numpy/reference/generated/numpy.fft.rfftfreq.html
    """
    # spacing between frequency bins
    spacing = epochs.info["sfreq"] / epochs.get_data().shape[-1]

    low_cutoff = tuple(band[0] for band in kwargs["freqs"])
    high_cutoff = tuple(
        band[1] - spacing if band != max(kwargs["freqs"]) else band[1]
        for band in kwargs["freqs"]
    )

    epochs_csd = compute_surface_laplacian(epochs=epochs, verbose=False)
    con, _, _, _, _ = spectral_connectivity(
        data=epochs_csd,
        method=kwargs["method"],
        sfreq=epochs.info["sfreq"],
        mode=kwargs["mode"],
        fmin=low_cutoff,
        fmax=high_cutoff,
        faverage=kwargs["faverage"],
        n_jobs=kwargs["n_jobs"],
        verbose=True,
    )

    # from shape of (n_signals, n_signals, n_freqs) to
    # (n_freqs, n_signals, n_signals)
    con = np.transpose(con, (2, 0, 1))
    con = abs(con)

    return con


def compute_surface_laplacian(epochs: Epochs, verbose: bool = True) -> Epochs:
    """
    Performs the surface Laplacian transform on the Epochs instance
    For more information about the transform parameters please refer to
    Cohen, M. X. (2014). Analyzing neural time series data: theory and practice
    . MIT press. For more information about this function in particular,
    visit the MNE documentation at:
    https://mne.tools/dev/generated/mne.preprocessing.compute_current_source_density.html
    Parameters
    ----------
    epochs: the epochs to be transformed
    verbose: whether to visualize the power spectral densities before and after
    the Laplacian transform

    Returns
    -------
    Raw instance
    """

    epochs_csd = compute_current_source_density(epochs.copy())

    if verbose:
        fig, axes_subplot = plt.subplots(
            nrows=2, ncols=1, sharex="all", sharey="all", dpi=220
        )

        epochs.plot_psd(ax=axes_subplot[0], show=False, fmax=60)
        epochs_csd.plot_psd(ax=axes_subplot[1], show=False, fmax=60)
        fig.show()

    return epochs_csd



"""
This module contains functions that can be used to perform time-frequency analysis on
EEG/MEG data with MNE-Python.
"""
import os
import re

import numpy as np

from mne import Epochs
from mne.time_frequency import tfr_morlet, AverageTFR
from mne.utils import logger
from yasa import irasa


def compute_power(epochs: Epochs, config: dict) -> AverageTFR:
    """
    Computes Time-Frequency Representation (TFR) using complex Morlet wavelets
    averaged over epochs. Power is written to an HDF5 file.
    Parameters
    ----------
    epochs
    config
    """
    fmin = config["morlet"]["fmin"]
    fmax = config["morlet"]["fmax"]
    step = config["morlet"]["step"]
    freqs = np.logspace(*np.log10([fmin, fmax]), num=step)
    n_cycles = freqs / 2.0

    power = tfr_morlet(
        epochs.average() if config["is_evoked"] else epochs.copy(),
        freqs=freqs,
        n_cycles=n_cycles,
        use_fft=False,
        return_itc=False,
        decim=config["morlet"]["decim"],
        average=True,
        n_jobs=-1,
    )

    return power


def save_to_hdf5(power: AverageTFR):
    if not os.path.exists(power.comment.parent):
        os.makedirs(power.comment.parent)

    power.comment = str(power.comment)
    # replace floats in file name with integers
    floats = re.findall("[-+]?\d*\.\d+", power.comment)
    if floats:
        for num in floats:
            power.comment = power.comment.replace(num, str(int(float(num))))
    file_path_with_extension = f"{power.comment}_power-tfr.h5"

    logger.info(f"Saving power at {file_path_with_extension} ...")
    power.save(fname=str(file_path_with_extension), overwrite=True)
    logger.info("[FINISHED]")


def average_power_into_frequency_bands(power: AverageTFR, config: dict) -> AverageTFR:
    """
    Computes average power in specific frequency ranges defined in config
    Parameters
    ----------
    power
    config

    Returns
    -------

    """
    band_values = config.values()
    band_power_data_arr = []

    for band in band_values:
        band_filter = np.logical_and(
            power.freqs >= float(band[0]), power.freqs < float(band[1])
        )
        if int(max(power.freqs)) == int(max(band_values)[1]):
            band_filter[-1] = True
        band_data = power.data[:, band_filter, :].mean(axis=1)
        band_power_data_arr.append(band_data[np.newaxis, :])

    band_power_data = np.concatenate(band_power_data_arr, axis=0)
    band_power_data = np.transpose(band_power_data, (1, 0, 2))

    band_freqs = np.asarray([band[0] for band in band_values])

    band_power = power.copy()
    band_power.data = band_power_data
    band_power.freqs = band_freqs
    band_power.comment = power.comment + "_band_average"

    return band_power


def compute_power_irasa_method(epochs: Epochs, bands: tuple):
    """
    Computes power by separating the aperiodic (= fractal, or 1/f) and oscillatory
    component of the power spectra of EEG data using the IRASA method.
    https://raphaelvallat.com/yasa/build/html/generated/yasa.irasa.html
    Parameters
    ----------
    epochs
    bands

    Returns
    -------

    """
    # transform epochs to continuous data
    # to shape of (n_channels, n_epochs, n_times)
    data = np.transpose(epochs.get_data(), (1, 0, 2))
    # reshape to (n_channels, n_epochs * n_times) continuous data
    data = data.reshape((data.shape[0], data.shape[1] * data.shape[2]))

    freqs, _, psd_oscillatory, _ = irasa(
        data=data,
        sf=epochs.info["sfreq"],
        ch_names=epochs.info["ch_names"],
        band=bands,
        return_fit=True,
        win_sec=2 * (1 / bands[0]),
    )

    return freqs, psd_oscillatory



settings = dict(
    bandpass_filter=dict(low_freq=0.5, high_freq=45),
    epochs=dict(start_time=0.0, end_time=1.0, duration=1),
    ica=dict(n_components=32, method="picard", decim=None),
    autoreject=dict(threshold=0.15),
)
